# -*- coding: UTF-8 -*-
# Python
"""
08-09-2023
@author: jeremy auclair

Main script to plot ETR and NDVI GIFs
"""

if __name__ == '__main__':
    
    import sys, os  # system management
    currentdir = os.path.dirname(os.path.realpath(__file__))
    parentdir = os.path.dirname(currentdir)
    sys.path.insert(0, os.path.dirname(parentdir))
    
    
    import argparse  # for parsing arguments
    import xarray as xr  # to manage dataset
    import numpy as np  # vectorized math
    import matplotlib.pyplot as plt  # to plot
    from matplotlib import rcParams  # plot parameters
    import warnings  # to turn off warnings
    from p_tqdm import p_umap  # for multiprocessing with progress bars
    from PIL import Image  # to generate GIF
    import contextlib  # to generate GIF
    import gc  # garbage collector module
    import matplotlib  # to prevent showing graphs
    matplotlib.use('Agg')  # to prevent showing graphs
    from modspa_pixel.config.config import config  # to import config file
    from modspa_pixel.postprocessing.output_toolbox import get_continuous_cmap
    from modspa_pixel.source.code_toolbox import format_duration  # to print memory requirements
    from time import time  # to time code excecution
    
    # Get start time of script
    t0 = time()
    
    ## Arguments to run the script
    parser = argparse.ArgumentParser()
    parser.add_argument("variable", type = str, choices = ['NDVI', 'ETR', 'DP', 'Irr', 'E', 'Tr', 'SWCe', 'SWCr'],  help = "Variables to generate a daily GIF from, choices = 0 = NDVI, 1 = ETR, 2 = DP, 3 = Irr, 4 = E, 5 = Tr, 6 = SWCe, 7 = SWCr")
    args = parser.parse_args()
    variable = args.variable
    
    # Xarray parameters
    xr.set_options(keep_attrs = True)  # always keep attributes even after operations

    # Turn off warnings
    np.seterr(divide='ignore', invalid='ignore')
    warnings.simplefilter("ignore", category = RuntimeWarning)

    # Parameters for matplotlib
    plt.style.use('default')
    rcParams['mathtext.fontset'] = 'stix'
    rcParams['font.family'] = 'STIXGeneral'
    rcParams.update({'font.size': 15})
    
    # Custom colormap for ndvi
    hex_list = ['ffffff', 'ce7e45', 'df923d', 'f1b555', 'fcd163', '99b718',
                '74a901', '66a000', '529400', '3e8601', '207401', '056201',
                '004c00', '023b01', '012e01', '011d01', '011301']
    cmap_ndvi = get_continuous_cmap(hex_list)
    
    # Declare paths
    config_file = parentdir + os.sep + 'config' + os.sep + 'config_modspa.json'
    
    # Open config file and load parameters
    config_params = config(config_file)
    
    # General parameters
    max_cpu = config_params.max_cpu
    run_name = config_params.run_name
    preferred_provider = config_params.preferred_provider
    start_date = config_params.start_date
    end_date = config_params.end_date

    # Load paths
    output_path = config_params.output_path
    download_path = config_params.download_path
    
    # Land Cover path
    land_cover_path = config_params.land_cover_path
    
    # Output path
    output_save_path = output_path + os.sep + run_name + os.sep + run_name + '_outputs.nc'
    
    # NDVI input path
    if preferred_provider == 'copernicus':
        ndvi_path = download_path + os.sep + 'SCIHUB' + os.sep + 'NDVI' + os.sep + run_name + '_NDVI_cube_' + start_date + '_' + end_date + '.nc'
    else:
        ndvi_path = download_path + os.sep + 'THEIA' + os.sep + 'NDVI' + os.sep + run_name + '_NDVI_cube_' + start_date + '_' + end_date + '.nc'

    # Open output dataset
    output_dataset = xr.open_dataset(output_save_path, chunks = {'x': 2000, 'y': 2000, 'time': 4})
    dates = output_dataset.time.values

    # Open Land Cover raster
    land_cover = xr.open_dataarray(land_cover_path)
    
    # Save path
    output_gif_path = output_path  + os.sep + run_name + os.sep + 'GIF'
    if not os.path.exists(output_gif_path): os.mkdir(output_gif_path)
    gif_var_path = output_gif_path + os.sep + variable + '.gif'
    
    # Define plot function
    def plot_image(args: tuple) -> str:
        """
        Generate a matplotlib image and save it. Used as a frame for
        GIF generation.

        Arguments (packed in args: ``tuple``)
        =====================================

        1. day: ``int``
            index of current day
        2. daily_array: ``xr.DataArray``
            DataArray of current day
        3. cmap: ``matplotlib.colors.Colormap``
            colormap to plot
        4. title: ``str``
            title to show for current figure
        5. variable: ``str``
            name of variable (for save file)
        6. colorbar_label: ``str``
            colorbar label
        6. vmin: ``float``
            minimum value of data
        7. vmax: ``float``
            maximum value of data
        8. output_gif_path: ``float``
            directory in which to save images

        Returns
        =======

        1. figure_path : ``str``
            path of saved image
        """
        
        # Unpack args
        day, daily_array, cmap, title, variable, colorbar_label, vmin, vmax, output_gif_path = args
        
        # Define plot
        fig = plt.figure(figsize = (9,7))
        ax = fig.gca()
        
        # Plot data
        im = daily_array.plot.imshow(cmap = cmap, add_colorbar = False, vmin = vmin, vmax = vmax)
        ax.set_aspect('equal', 'box')
        plt.title(title);
        fig.tight_layout()
        
        # Manage colorbar
        cax = fig.add_axes([ax.get_position().x1+0.01,ax.get_position().y0,0.03,ax.get_position().height])
        cbar = plt.colorbar(im, cax = cax, format = "%.1f")
        cbar.set_label(label = colorbar_label, labelpad = 10)
        
        # Save image
        figure_path = output_gif_path + os.sep + f'{day+1:03d}_' + variable + '_.png'
        fig.savefig(figure_path, dpi = 100, bbox_inches='tight')
        
        # Release memory
        plt.close()
        fig.clf()
        del fig, ax, im # delete the objects
        gc.collect()  # collect garbage
        
        return figure_path

    
    if variable == 'ETR':
        # Calculate ETR
        output_dataset['ETR'] = (output_dataset.E + output_dataset.Tr) / 1000
        output_dataset['ETR'].values  = np.where(land_cover == 1, np.NaN, output_dataset['ETR'])
        output_dataset['ETR'].attrs['standard_name'] = 'E + Tr'
    
        # Generate arguments for ETR GIF
        vmin = 0.1
        vmax = 3.5
        args = [(day, output_dataset.ETR.isel(time = day), 'YlGnBu', 'Evapotranspiration -- '  + np.datetime_as_string(dates[day], unit='D'), 'ETR', 'E + Tr [mm]', vmin, vmax, output_gif_path) for day in range(len(dates))]

        # Generate frames
        print('\nGenerating frames for ETR GIF\n')
        results = p_umap(plot_image, args, **{"num_cpus": max_cpu})
        results.sort()

        # Close dataset to release memory
        output_dataset.close()
        
        # Generate GIF for ETR
        print('\nGenerating ETR GIF\n')
        with contextlib.ExitStack() as stack:

            # lazily load images
            imgs = (stack.enter_context(Image.open(f)) for f in results)

            # extract  first image from iterator
            img = next(imgs)

            # Save GIF
            img.save(fp = gif_var_path, format = 'GIF', append_images = imgs, save_all = True, duration = 100, loop = 2)

        # Remove unecessary images
        for file in results:
            os.remove(file)
    
    elif variable == 'NDVI':
        
        # Open NDVI dataset
        NDVI_dataset = xr.open_dataset(ndvi_path, chunks = {'x': 2000, 'y': 2000, 'time': 1})
        dates = NDVI_dataset.time.values
            
        # Generate arguments for NDVI GIF
        vmin = 0
        vmax = 1
        args = [(day, NDVI_dataset.NDVI.isel(time = day)/255, cmap_ndvi, 'NDVI -- '  + np.datetime_as_string(dates[day], unit='D'), 'NDVI', 'NDVI', vmin, vmax, output_gif_path) for day in range(len(dates))]

        # Generate frames
        print('\nGenerating frames for NDVI GIF\n')
        results = p_umap(plot_image, args, **{"num_cpus": max_cpu})
        results.sort()
        
        # Close dataset to release memory
        NDVI_dataset.close()
        
        # Generate GIF for NDVI
        print('\nGenerating NDVI GIF\n')
        with contextlib.ExitStack() as stack:

            # lazily load images
            imgs = (stack.enter_context(Image.open(f)) for f in results)

            # extract  first image from iterator
            img = next(imgs)
            
            # Save GIF
            img.save(fp = gif_var_path, format = 'GIF', append_images = imgs, save_all = True, duration = 100, loop = 2)

        # Remove unecessary images
        for file in results:
            os.remove(file)
    
    elif variable in ['E', 'Tr', 'SWCe', 'SWCr']:
        
        # Mask no_sim data
        output_dataset[variable].values  = np.where(land_cover == 1, np.NaN, output_dataset[variable])
        
        # Generate arguments for variable GIF
        if variable in ['E', 'Tr']:
            vmin = 0.1
            vmax = 3
            if variable == 'E': variable_name = 'Evaporation'
            else: variable_name = 'Transpiration'
        else:
            vmin = 0
            vmax = 1
            if variable == 'SWCe': variable_name = 'Soil Water Content evaporative layer'
            else: variable_name = 'Soil Water Content root layer'
        args = [(day, (output_dataset[variable].isel(time = day) / 1000), 'YlGnBu', variable_name + ' -- '  + np.datetime_as_string(dates[day], unit='D'), variable, variable + ' [mm]', vmin, vmax, output_gif_path) for day in range(len(dates))]

        # Generate frames
        print(f'\nGenerating frames for {variable} GIF\n')
        results = p_umap(plot_image, args, **{"num_cpus": max_cpu})
        results.sort()

        # Close dataset to release memory
        output_dataset.close()
        
        # Generate GIF for ETR
        print(f'\nGenerating {variable} GIF\n')
        with contextlib.ExitStack() as stack:

            # lazily load images
            imgs = (stack.enter_context(Image.open(f)) for f in results)

            # extract  first image from iterator
            img = next(imgs)

            # Save GIF
            img.save(fp = gif_var_path, format = 'GIF', append_images = imgs, save_all = True, duration = 100, loop = 2)

        # Remove unecessary images
        for file in results:
            os.remove(file)
    
    else:
        
        # Mask no_sim data
        output_dataset[variable].values  = np.where(land_cover == 1, np.NaN, output_dataset[variable])
        
        # Generate arguments for variable GIF
        if variable == 'DP': variable_name = 'Deep Percolation'
        else: variable_name = 'Irrigation'
        vmin = output_dataset[variable].min().values
        vmax = output_dataset[variable].max().values
        args = [(day, (output_dataset[variable].isel(time = day) / 100), 'YlGnBu', variable_name + ' -- '  + np.datetime_as_string(dates[day], unit='D'), variable, variable + ' [mm]', vmin, vmax, output_gif_path) for day in range(len(dates))]

        # Generate frames
        print(f'\nGenerating frames for {variable} GIF\n')
        results = p_umap(plot_image, args, **{"num_cpus": max_cpu})
        results.sort()

        # Close dataset to release memory
        output_dataset.close()
        
        # Generate GIF for ETR
        print(f'\nGenerating {variable} GIF\n')
        with contextlib.ExitStack() as stack:

            # lazily load images
            imgs = (stack.enter_context(Image.open(f)) for f in results)

            # extract  first image from iterator
            img = next(imgs)

            # Save GIF
            img.save(fp = gif_var_path, format = 'GIF', append_images = imgs, save_all = True, duration = 100, loop = 2)

        # Remove unecessary images
        for file in results:
            os.remove(file)
    
    # Print formatted runtime
    format_duration(time() - t0)